@@ -1,5 +1,6 @@
 #!/usr/bin/python -u
-# Copyright (c) 2004 Danilo Segan <danilo@kvota.net>.
+# -*- encoding: utf-8 -*-
+# Copyright (c) 2004, 2005, 2006 Danilo Šegan <danilo@gnome.org>.
 #
 # This file is part of xml2po.
 #
@@ -36,6 +37,27 @@ import gettext
 import os
 import re
 
+class NoneTranslations:
+    def gettext(self, message):
+        return None
+
+    def lgettext(self, message):
+        return None
+
+    def ngettext(self, msgid1, msgid2, n):
+        return None
+
+    def lngettext(self, msgid1, msgid2, n):
+        return None
+
+    def ugettext(self, message):
+        return None
+
+    def ungettext(self, msgid1, msgid2, n):
+        return None
+
+
+
 class MessageOutput:
     def __init__(self, with_translations = 0):
         self.messages = []
@@ -97,7 +119,7 @@ msgstr ""
 
     def outputAll(self, out):
         self.outputHeader(out)
-        
+
         for k in self.messages:
             if k in self.comments:
                 out.write("#. %s\n" % (self.comments[k].replace("\n","\n#. ")))
@@ -112,17 +134,23 @@ msgstr ""
             if self.do_translations:
                 if len(self.translations)>0:
                     translation = self.translations.pop(0)
+            if translation == k:
+                translation = ""
             out.write("msgstr \"%s\"\n\n" % (translation))
 
 
 def normalizeNode(node):
+    #print >>sys.stderr, "<%s> (%s) [%s]" % (node.name, node.type, node.serialize('utf-8'))
     if not node:
         return
     elif isSpacePreserveNode(node):
         return
     elif node.isText():
         if node.isBlankNode():
-            node.setContent('')
+            if expand_entities or ( not (node.prev and not node.prev.isBlankNode()
+                                         and node.next and not node.next.isBlankNode()) ):
+                #print >>sys.stderr, "BLANK"
+                node.setContent('')
         else:
             node.setContent(re.sub('\s+',' ', node.content))
 
@@ -167,7 +195,7 @@ def normalizeString(text, ignorewhitespace = 1):
 
     result = re.sub('^ ','', result)
     result = re.sub(' $','', result)
-    
+
     return result
 
 def stringForEntity(node):
@@ -213,16 +241,36 @@ def getTranslation(text, spacepreserve = 0):
     text should be a string to look for, spacepreserve set to 1
     when spaces should be preserved.
     """
+    #print >>sys.stderr,"getTranslation('%s')" % (text.encode('utf-8'))
     text = normalizeString(text, not spacepreserve)
     if (text.strip() == ''):
         return text
-    file = open(mofile, "rb")
-    if file:
-        gt = gettext.GNUTranslations(file)
-        if gt:
-            return gt.ugettext(text.decode('utf-8'))
+    global gt
+    if gt:
+        res = gt.ugettext(text.decode('utf-8'))
+        return res
+
     return text
 
+def myAttributeSerialize(node):
+    result = ''
+    if node.children:
+        child = node.children
+        while child:
+            if child.type=='text':
+                result += doc.encodeEntitiesReentrant(child.content)
+            elif child.type=='entity_ref':
+                if not expand_entities:
+                    result += '&' + child.name + ';'
+                else:
+                    result += child.content.decode('utf-8')
+            else:
+                result += myAttributeSerialize(child)
+            child = child.next
+    else:
+        result = node.serialize('utf-8')
+    return result
+
 def startTagForNode(node):
     if not node:
         return 0
@@ -232,17 +280,20 @@ def startTagForNode(node):
     if node.properties:
         for p in node.properties:
             if p.type == 'attribute':
-                # FIXME: This part sucks
-                params += p.serialize('utf-8')
+                try:
+                    nsprop = p.ns().name + ":" + p.name
+                except:
+                    nsprop = p.name
+                params += " %s=\"%s\"" % (nsprop, myAttributeSerialize(p))
     return result+params
-        
+
 def endTagForNode(node):
     if not node:
         return 0
 
     result = node.name
     return result
-        
+
 def isFinalNode(node):
     if automatic:
         auto = autoNodeIsFinal(node)
@@ -260,7 +311,7 @@ def isFinalNode(node):
         final_children = 1
         child = node.children
         while child and final_children:
-            if not isFinalNode(child):
+            if not child.isBlankNode() and child.type != 'comment' and not isFinalNode(child):
                 final_children = 0
             child = child.next
         if final_children:
@@ -302,32 +353,42 @@ def getCommentForNode(node):
     else:
         return None
 
+def replaceAttributeContentsWithText(node,text):
+    node.setContent(text)
 
 def replaceNodeContentsWithText(node,text):
     """Replaces all subnodes of a node with contents of text treated as XML."""
-    if node.children:
-        starttag = node.name #startTagForNode(node)
-        endtag = endTagForNode(node)
-        try:
-            # Lets add document DTD so entities are resolved
-            dtd = doc.intSubset()
-            tmp = ''
-            if expand_entities: # FIXME: we get a "Segmentation fault" in libxml2.parseMemory() when we include DTD otherwise
-                tmp = dtd.serialize('utf-8')
-            tmp = tmp + '<%s>%s</%s>' % (starttag, text, endtag)
-        except:
-            tmp = '<%s>%s</%s>' % (starttag, text, endtag)
 
+    if node.children:
+        starttag = startTagForNode(node)
+        endtag = endTagForNode(node)
+
+        # Lets add document DTD so entities are resolved
+        tmp = '<?xml version="1.0" encoding="utf-8" ?>'
         try:
-            ctxt = libxml2.createDocParserCtxt(tmp.encode('utf-8'))
+            dtd = doc.intSubset()
+            tmp = tmp + dtd.serialize('utf-8')
+        except libxml2.treeError:
+            pass
+
+        content = '<%s>%s</%s>' % (starttag, text, endtag)
+        tmp = tmp + content.encode('utf-8')
+
+        newnode = None
+        try:
+            ctxt = libxml2.createDocParserCtxt(tmp)
             ctxt.replaceEntities(0)
             ctxt.parseDocument()
             newnode = ctxt.doc()
         except:
+            pass
+
+        if not newnode:
             print >> sys.stderr, """Error while parsing translation as XML:\n"%s"\n""" % (text.encode('utf-8'))
             return
 
         newelem = newnode.getRootElement()
+
         if newelem and newelem.children:
             free = node.children
             while free:
@@ -335,7 +396,12 @@ def replaceNodeContentsWithText(node,text):
                 free.unlinkNode()
                 free = next
 
-            node.addChildList(newelem.children)
+            if node:
+                copy = newelem.copyNodeList()
+                next = node.next
+                node.replaceNode(newelem.copyNodeList())
+                node.next = next
+
         else:
             # In practice, this happens with tags such as "<para>    </para>" (only whitespace in between)
             pass
@@ -344,9 +410,15 @@ def replaceNodeContentsWithText(node,text):
 
 def autoNodeIsFinal(node):
     """Returns 1 if node is text node, contains non-whitespace text nodes or entities."""
-    final = 0
+    if hasattr(node, '__autofinal__'):
+        return node.__autofinal__
+    if node.name in ignored_tags:
+        node.__autofinal__ = 0
+        return 0
     if node.isText() and node.content.strip()!='':
+        node.__autofinal__ = 1
         return 1
+    final = 0
     child = node.children
     while child:
         if child.type in ['text'] and  child.content.strip()!='':
@@ -354,15 +426,20 @@ def autoNodeIsFinal(node):
             break
         child = child.next
 
+    node.__autofinal__ = final
     return final
 
 
-def worthOutputting(node):
+def worthOutputting(node, noauto = 0):
     """Returns 1 if node is "worth outputting", otherwise 0.
 
     Node is "worth outputting", if none of the parents
     isFinalNode, and it contains non-blank text and entities.
     """
+    if noauto and hasattr(node, '__worth__'):
+        return node.__worth__
+    elif not noauto and hasattr(node, '__autoworth__'):
+        return node.__autoworth__
     worth = 1
     parent = node.parent
     final = isFinalNode(node) and node.name not in ignored_tags
@@ -374,13 +451,37 @@ def worthOutputting(node):
             break
         parent = parent.parent
     if not worth:
+        node.__worth__ = 0
         return 0
 
-    return autoNodeIsFinal(node)
-    
+    if noauto:
+        node.__worth__ = worth
+        return worth
+    else:
+        node.__autoworth__ = autoNodeIsFinal(node)
+        return node.__autoworth__
+
+def processAttribute(node, attr):
+    if not node or not attr or not worthOutputting(node=node, noauto=1):
+        return
+
+    outtxt = attr.content
+    if mode=='merge':
+        translation = getTranslation(outtxt, 0)
+        replaceAttributeContentsWithText(attr, translation.encode('utf-8'))
+    else:
+        msg.outputMessage(outtxt, node.lineNo(),  "", 0,
+                          node.name + ":" + attr.name)
+
 def processElementTag(node, replacements, restart = 0):
     """Process node with node.type == 'element'."""
     if node.type == 'element':
+        # Translate attributes if needed
+        if node.properties and len(treated_attributes):
+            for p in node.properties:
+                if p.name in treated_attributes:
+                    processAttribute(node, p)
+
         outtxt = ''
         if restart:
             myrepl = []
@@ -407,17 +508,23 @@ def processElementTag(node, replacements, restart = 0):
             translation = getTranslation(outtxt, isSpacePreserveNode(node))
         else:
             translation = outtxt
+
         starttag = startTagForNode(node)
         endtag = endTagForNode(node)
 
-        if restart or worthOutputting(node):
+        worth = worthOutputting(node)
+        if not translation:
+            translation = outtxt.decode('utf-8')
+            if worth and mark_untranslated: node.setLang('C')
+
+        if restart or worth:
             i = 0
             while i < len(myrepl):
                 replacement = '<%s>%s</%s>' % (myrepl[i][0], myrepl[i][3], myrepl[i][2])
                 i += 1
                 translation = translation.replace('<placeholder-%d/>' % (i), replacement)
 
-            if worthOutputting(node):
+            if worth:
                 if mode == 'merge':
                     replaceNodeContentsWithText(node, translation)
                 else:
@@ -483,7 +590,7 @@ def doSerialize(node):
             child = child.next
         return outtxt
 
-    
+
 def read_finaltags(filelist):
     if CurrentXmlMode:
         return CurrentXmlMode.getFinalTags()
@@ -501,11 +608,19 @@ def read_ignoredtags(filelist):
                     'varlistentry' ]
         return defaults
 
+def read_treatedattributes(filelist):
+    if CurrentXmlMode:
+        return CurrentXmlMode.getTreatedAttributes()
+    else:
+        return []
+
+
 def tryToUpdate(allargs, lang):
     # Remove "-u" and "--update-translation"
+    print >>sys.stderr, "OVDI!"
     command = allargs[0]
     args = allargs[1:]
-    opts, args = getopt.getopt(args, 'avhmket:o:p:u:',
+    opts, args = getopt.getopt(args, 'avhm:ket:o:p:u:',
                                ['automatic-tags','version', 'help', 'keep-entities', 'extract-all-entities', 'merge', 'translation=',
                                 'output=', 'po-file=', 'update-translation=' ])
     for opt, arg in opts:
@@ -578,6 +693,7 @@ default_mode = 'docbook'
 filename = ''
 origxml = ''
 mofile = ''
+gt = None
 ultimate = [ ]
 ignored = [ ]
 filenames = [ ]
@@ -586,6 +702,7 @@ translationlanguage = ''
 mode = 'pot' # 'pot' or 'merge'
 automatic = 0
 expand_entities = 1
+mark_untranslated = 0
 expand_all_entities = 0
 
 output  = '-' # this means to stdout
@@ -611,7 +728,10 @@ OPTIONS may be some of:
     -r    --reuse=FILE         Specify translated XML file with the same structure
     -t    --translation=FILE   Specify MO file containing translation, and merge
     -u    --update-translation=LANG.po   Updates a PO file using msgmerge program
+
     -l    --language=LANG      Set language of the translation to LANG
+          --mark-untranslated  Set 'xml:lang="C"' on untranslated tags
+
     -v    --version            Output version of the xml2po program
 
     -h    --help               Output this message
@@ -633,7 +753,7 @@ if len(sys.argv) < 2: usage()
 args = sys.argv[1:]
 try: opts, args = getopt.getopt(args, 'avhkem:t:o:p:u:r:l:',
                            ['automatic-tags','version', 'help', 'keep-entities', 'expand-all-entities', 'mode=', 'translation=',
-                            'output=', 'po-file=', 'update-translation=', 'reuse=', 'language=' ])
+                            'output=', 'po-file=', 'update-translation=', 'reuse=', 'language=', 'mark-untranslated' ])
 except getopt.GetoptError: usage(True)
 
 for opt, arg in opts:
@@ -643,6 +763,8 @@ for opt, arg in opts:
         automatic = 1
     elif opt in ('-k', '--keep-entities'):
         expand_entities = 0
+    elif opt in ('--mark-untranslated',):
+        mark_untranslated = 1
     elif opt in ('-e', '--expand-all-entities'):
         expand_all_entities = 1
     elif opt in ('-l', '--language'):
@@ -688,8 +810,18 @@ if mode=='merge' and mofile=='':
     print >> sys.stderr, "Error: You must specify MO file when merging translations."
     sys.exit(3)
 
+if mofile:
+    try:
+        file = open(mofile, "rb")
+
+        gt = gettext.GNUTranslations(file)
+        gt.add_fallback(NoneTranslations())
+    except:
+        print >> sys.stderr, "Can't open MO file '%s'." % (mofile)
+
 ultimate_tags = read_finaltags(ultimate)
 ignored_tags = read_ignoredtags(ignored)
+treated_attributes = read_treatedattributes(ignored)
 
 # I'm not particularly happy about making any of these global,
 # but I don't want to bother too much with it right now
